@@ -1,17 +1,18 @@
 package JIRA::REST;
-# ABSTRACT: Thin wrapper around JIRA's REST API
-$JIRA::REST::VERSION = '0.017';
-use 5.008_008;
+# ABSTRACT: Thin wrapper around Jira's REST API
+$JIRA::REST::VERSION = '0.021';
+use 5.016;
 use utf8;
-use strict;
 use warnings;
 
 use Carp;
 use URI;
+use Encode;
 use MIME::Base64;
 use URI::Escape;
-use JSON;
+use JSON 2.23;
 use REST::Client;
+use HTTP::CookieJar::LWP;
 
 sub new {
     my $class = shift; # this always has to come first!
@@ -46,6 +47,12 @@ sub new {
         $api = $1;
         $args{url}->path($path);
     }
+    # Strip trailing slashes from $path. For some reason they cause 404
+    # errors down the road.
+    if ($path =~ s:/+$::) {
+        $args{url}->path($path);
+    }
+
     unless ($args{anonymous}) {
         # If username and password are not set we try to lookup the credentials
         if (! defined $args{username} || ! defined $args{password}) {
@@ -55,12 +62,12 @@ sub new {
 
         foreach (qw/username password/) {
             croak __PACKAGE__ . "::new: '$_' argument must be a non-empty string.\n"
-                unless defined $args{$_} && ! ref $args{$_} && length $args{$_};
+                if ! defined $args{$_} || ref $args{$_} || length $args{$_} == 0;
         }
     }
 
     for ($args{rest_client_config}) {
-        $_ = {} unless defined;
+        $_ //= {};
         croak __PACKAGE__ . "::new: 'rest_client_config' argument must be a hash reference.\n"
             unless defined && ref && ref eq 'HASH';
     }
@@ -69,7 +76,7 @@ sub new {
     # This is deprecated since v0.017
     if (my $proxy = delete $args{rest_client_config}{proxy}) {
         carp __PACKAGE__ . "::new: passing 'proxy' in the 'rest_client_config' hash is deprecated. Please, use the corresponding argument instead.\n";
-        $args{proxy} = $proxy unless defined $args{proxy};
+        $args{proxy} //= $proxy;
     }
 
     my $rest = REST::Client->new($args{rest_client_config});
@@ -80,9 +87,9 @@ sub new {
     # Follow redirects/authentication by default
     $rest->setFollow(1);
 
-    # Since JIRA doesn't send an authentication challenge, we force the
+    # Since Jira doesn't send an authentication challenge, we force the
     # sending of the authentication header.
-    $rest->addHeader(Authorization => 'Basic ' . encode_base64("$args{username}:$args{password}"))
+    $rest->addHeader(Authorization => 'Basic ' . encode_base64("$args{username}:$args{password}", ''))
         unless $args{anonymous};
 
     for my $ua ($rest->getUseragent) {
@@ -94,15 +101,35 @@ sub new {
 
         # Turn off SSL verification if requested
         $ua->ssl_opts(SSL_verify_mode => 0, verify_hostname => 0) if $args{ssl_verify_none};
+
+        # Configure a cookie_jar so that we can send Cookie headers
+        $ua->cookie_jar(HTTP::CookieJar::LWP->new());
     }
 
     return bless {
         rest => $rest,
         json => JSON->new->utf8->allow_nonref,
         api  => $api,
+        args => \%args,
     } => $class;
 }
 
+sub new_session {
+    my ($class, @args) = @_;
+    my $jira = $class->new(@args);
+    $jira->{_session} = $jira->POST('/rest/auth/1/session', undef, {
+        username => $jira->{args}{username},
+        password => $jira->{args}{password},
+    });
+    return $jira;
+}
+
+sub DESTROY {
+    my $self = shift;
+    $self->DELETE('/rest/auth/1/session') if exists $self->{_session};
+    return;
+}
+
 sub _search_for_credentials {
     my ($URL, $username) = @_;
     my (@errors, $password);
@@ -173,7 +200,7 @@ sub _error {
     } elsif ($type =~ m:application/json:) {
         my $error = $self->{json}->decode($content);
         if (ref $error eq 'HASH') {
-            # JIRA errors may be laid out in all sorts of ways. You have to
+            # Jira errors may be laid out in all sorts of ways. You have to
             # look them up from the scant documentation at
             # https://docs.atlassian.com/jira/REST/latest/.
 
@@ -190,6 +217,9 @@ sub _error {
             if (my $errors = $error->{errors}) {
                 $msg .= "- [$_] $errors->{$_}\n" foreach sort keys %$errors;
             }
+
+            # some give us a single message in 'errorMessage'
+            $msg .= $error->{errorMessage} . qq{\n} if $error->{errorMessage};
         } else {
             $msg .= $content;
         }
@@ -231,8 +261,9 @@ sub _content {
 sub _build_path {
     my ($self, $path, $query) = @_;
 
-    # Prefix $path with the default API prefix unless it already specifies one
-    $path = $self->{api} . $path unless $path =~ m:^/rest/:;
+    # Prefix $path with the default API prefix unless it already specifies
+    # one or it's an absolute URL.
+    $path = $self->{api} . $path unless $path =~ m@^(?:/rest/|(?i)https?:)@;
 
     if (defined $query) {
         croak $self->_error("The QUERY argument must be a hash reference.")
@@ -268,8 +299,7 @@ sub PUT {
     $path = $self->_build_path($path, $query);
 
     $headers                   ||= {};
-    $headers->{'Content-Type'}   = 'application/json;charset=UTF-8'
-        unless defined $headers->{'Content-Type'};
+    $headers->{'Content-Type'} //= 'application/json;charset=UTF-8';
 
     $self->{rest}->PUT($path, $self->{json}->encode($value), $headers);
 
@@ -285,8 +315,7 @@ sub POST {
     $path = $self->_build_path($path, $query);
 
     $headers                   ||= {};
-    $headers->{'Content-Type'}   = 'application/json;charset=UTF-8'
-        unless defined $headers->{'Content-Type'};
+    $headers->{'Content-Type'} //= 'application/json;charset=UTF-8';
 
     $self->{rest}->POST($path, $self->{json}->encode($value), $headers);
 
@@ -348,12 +377,14 @@ sub attach_files {
             %{$rest->{_headers}},
             'X-Atlassian-Token' => 'nocheck',
             'Content-Type'      => 'form-data',
-            'Content'           => [ file => [$file] ],
+            'Content'           => [ file => [$file, encode_utf8( $file )] ],
         );
 
         $response->is_success
             or croak $self->_error("attach_files($file): " . $response->status_line);
     }
+
+    return;
 }
 
 1;
@@ -366,18 +397,18 @@ __END__
 
 =head1 NAME
 
-JIRA::REST - Thin wrapper around JIRA's REST API
+JIRA::REST - Thin wrapper around Jira's REST API
 
 =head1 VERSION
 
-version 0.017
+version 0.021
 
 =head1 SYNOPSIS
 
     use JIRA::REST;
 
     my $jira = JIRA::REST->new({
-        URL      => 'https://jira.example.net',
+        url      => 'https://jira.example.net',
         username => 'myuser',
         password => 'mypass',
     });
@@ -423,44 +454,44 @@ version 0.017
 
 =head1 DESCRIPTION
 
-L<JIRA|http://www.atlassian.com/software/jira/> is a proprietary bug
+L<Jira|http://www.atlassian.com/software/jira/> is a proprietary bug
 tracking system from Atlassian.
 
-This module implements a very thin wrapper around JIRA's REST APIs:
+This module implements a very thin wrapper around Jira's REST APIs:
 
 =over
 
-=item * L<JIRA Core REST API|https://docs.atlassian.com/jira/REST/server/>
+=item * L<Jira Core REST API|https://docs.atlassian.com/software/jira/docs/api/REST/latest/>
 
-This rich API superseded the old L<JIRA SOAP
+This rich API superseded the old L<Jira SOAP
 API|http://docs.atlassian.com/software/jira/docs/api/rpc-jira-plugin/latest/com/atlassian/jira/rpc/soap/JiraSoapService.html>
-which isn't supported anymore as of JIRA version 7.
+which isn't supported anymore as of Jira version 7.
 
 The endpoints of this API have a path prefix of C</rest/api/VERSION>.
 
-=item * L<JIRA Service Desk REST API|https://docs.atlassian.com/jira-servicedesk/REST/server/>
+=item * L<Jira Service Desk REST API|https://docs.atlassian.com/jira-servicedesk/REST/server/>
 
-This API deals with the objects of the JIRA Service Desk application. Its
+This API deals with the objects of the Jira Service Desk application. Its
 endpoints have a path prefix of C</rest/servicedeskapi>.
 
-=item * L<JIRA Software REST API|https://docs.atlassian.com/jira-software/REST/server/>
+=item * L<Jira Software REST API|https://docs.atlassian.com/jira-software/REST/server/>
 
-This API deals with the objects of the JIRA Software application. Its
+This API deals with the objects of the Jira Software application. Its
 endpoints have a path prefix of C</rest/agile/VERSION>.
 
 =back
 
-=head1 CONSTRUCTOR
+=head1 CONSTRUCTORS
 
 =head2 new HASHREF
 
 =head2 new URL, USERNAME, PASSWORD, REST_CLIENT_CONFIG, ANONYMOUS, PROXY, SSL_VERIFY_NONE
 
-The constructor can take its arguments from a single hash reference or from
-a list of positional parameters. The first form is preferred because it lets
-you specify only the arguments you need. The second form forces you to pass
-undefined values if you need to pass a specific value to an argument further
-to the right.
+The default constructor can take its arguments from a single hash reference or
+from a list of positional parameters. The first form is preferred because it
+lets you specify only the arguments you need. The second form forces you to pass
+undefined values if you need to pass a specific value to an argument further to
+the right.
 
 The arguments are described below with the names which must be used as the
 hash keys:
@@ -469,7 +500,7 @@ hash keys:
 
 =item * B<url>
 
-A string or a URI object denoting the base URL of the JIRA server. This is a
+A string or a URI object denoting the base URL of the Jira server. This is a
 required argument.
 
 The REST methods described below all accept as a first argument the
@@ -497,7 +528,7 @@ during construction.
 
 =item * B<password>
 
-The username and password of a JIRA user to use for authentication.
+The username and password of a Jira user to use for authentication.
 
 If B<anonymous> is false then, if either B<username> or B<password> isn't
 defined the module looks them up in either the C<.netrc> file or via
@@ -518,7 +549,7 @@ As an extension, the hash reference also accepts one additional argument
 called B<proxy> that is an extension to the REST::Client configuration and
 will be removed from the hash before passing it on to the REST::Client
 constructor. However, this argument is deprecated since v0.017 and you
-should avoid it. Instead, use the following argument instead.
+should avoid it. Use the following argument instead.
 
 =item * B<proxy>
 
@@ -529,23 +560,36 @@ describing the fully qualified URL (including port) to your network proxy.
 
 Sets the C<SSL_verify_mode> and C<verify_hostname ssl> options on the
 underlying L<REST::Client>'s user agent to 0, thus disabling them. This
-allows access to JIRA servers that have self-signed certificates that don't
+allows access to Jira servers that have self-signed certificates that don't
 pass L<LWP::UserAgent>'s verification methods.
 
 =item * B<anonymous>
 
-Tells the module that you want to connect to the specified JIRA server with
-no username or password.  This way you can access public JIRA servers
+Tells the module that you want to connect to the specified Jira server with
+no username or password.  This way you can access public Jira servers
 without needing to authenticate.
 
 =back
 
+=head2 new_session OPTIONS
+
+This 'session' constructor first invokes the default constructor, passing to it
+all the options it receives. Then it makes a C<POST /rest/auth/1/session> to
+login to Jira, creating a user session.
+
+This is particularly useful when interacting with Jira Data Center, because it
+can use the session cookie to maintain affinity with one of the redundant
+servers.
+
+When created with this constructor, upon destruction the object makes a C<DELETE
+/rest/auth/1/session> to logout from Jira.
+
 =head1 REST METHODS
 
-JIRA's REST API documentation lists dozens of "resources" which can be
+Jira's REST API documentation lists dozens of "resources" which can be
 operated via the standard HTTP requests: GET, DELETE, PUT, and
 POST. JIRA::REST objects implement four methods called GET, DELETE,
-PUT, and POST to make it easier to invoke and get results from JIRA's
+PUT, and POST to make it easier to invoke and get results from Jira's
 REST endpoints.
 
 All four methods need two arguments:
@@ -559,7 +603,7 @@ fields, you pass C</rest/api/latest/field>, and in order to get SLA
 information about an issue you pass
 C</rest/servicedeskapi/request/$key/sla>.
 
-If you're using a method form JIRA Core REST API you may omit the prefix
+If you're using a method from Jira Core REST API you may omit the prefix
 C</rest/api/VERSION>. For example, to GET the list of all fields you may
 pass just C</field>.
 
@@ -591,7 +635,7 @@ usually is a hash reference. The value is encoded as a
 L<JSON|http://www.json.org/> string using the C<JSON::encode> method
 and sent with a Content-Type of C<application/json>.
 
-It's usually easy to infer from the JIRA REST API documentation which
+It's usually easy to infer from the Jira REST API documentation which
 kind of value you should pass to each resource.
 
 This argument is required.
@@ -664,7 +708,7 @@ Sets up an iterator for the search specified by the hash reference PARAMS.
 It must be called before calls to B<next_issue>.
 
 PARAMS must conform with the query parameters allowed for the
-C</rest/api/2/search> JIRA REST endpoint.
+C</rest/api/2/search> Jira REST endpoint.
 
 =head2 B<next_issue>
 
@@ -684,6 +728,33 @@ requires a specific content type encoding which is difficult to come up with
 just the C<REST::Client> interface. This utility method offers an easier
 interface to attach files to issues.
 
+=head1 PERL AND JIRA COMPATIBILITY POLICY
+
+Currently L<JIRA::REST> requires Perl 5.16 and supports Jira 7.0.
+
+We try to be compatible with the Perl native packages of the oldest L<Ubuntu
+LTS|https://www.ubuntu.com/info/release-end-of-life> and
+L<CentOS|https://wiki.centos.org/About/Product> Linux distributions still
+getting maintainance updates.
+
+  +-------------+-----------------------+------+
+  | End of Life | Distro                | Perl |
+  +-------------+-----------------------+------+
+  |   2021-04   | Ubuntu 16.04 (xenial) | 5.22 |
+  |   2023-04   | Ubuntu 18.04 (bionic) | 5.26 |
+  |   2024-07   | CentOS 7              | 5.16 |
+  |   2025-04   | Ubuntu 20.04 (focal ) | 5.30 |
+  |   2029-05   | CentOS 8              | 5.26 |
+  +-------------+-----------------------+------+
+
+As you can see, we're kept behind mostly by the slow pace of CentOS (actually,
+RHEL) releases.
+
+As for Jira, the policy is very lax. I (the author) only test L<JIRA::REST> on
+the Jira server installed in the company I work for, which is usually (but not
+always) at most one year older than the newest released version. I don't have
+yet an easy way to test it on different versions.
+
 =head1 SEE ALSO
 
 =over
@@ -692,9 +763,13 @@ interface to attach files to issues.
 
 JIRA::REST uses a REST::Client object to perform the low-level interactions.
 
+=item * C<JIRA::REST::OAuth>
+
+This module Sub Classes JIRA::REST providing OAuth 1.0 support.
+
 =item * C<JIRA::Client::REST>
 
-This is another module implementing JIRA's REST API using
+This is another module implementing Jira's REST API using
 L<SPORE|https://github.com/SPORE/specifications/blob/master/spore_description.pod>.
 I got a message from the author saying that he doesn't intend to keep
 it going.
@@ -711,7 +786,7 @@ Gustavo L. de M. Chaves <gnustavo@cpan.org>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2017 by CPqD <www.cpqd.com.br>.
+This software is copyright (c) 2021 by CPQD <www.cpqd.com.br>.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
